from typing import Any, Callable, Dict, List, TypeVar

from ..typing import DEFAULT_LIFECYCLE_NAMES, T_Dispatcher

T = TypeVar("T")
I = TypeVar("I")

LF_TEMPLATE = {i: list() for i in DEFAULT_LIFECYCLE_NAMES}


class DII_NestableIterable:
    __slots__ = ("iterable", "indexes")

    iterable: List
    indexes: List

    def __init__(self, iterable: List) -> None:
        self.iterable = iterable
        self.indexes = [[0, 0]]

    def __iter__(self) -> T_Dispatcher:
        dis_set_index, dis_index = self.indexes[-1]
        dis_set_index_offset = dis_set_index + (dis_set_index and 1)
        dis_index_offset = dis_index + (dis_index and 1)

        current_indexes = [dis_set_index_offset, dis_index_offset]
        self.indexes.append(current_indexes)

        for content in self.iterable[dis_set_index_offset:]:
            for i in content[dis_index_offset:]:
                yield i
                current_indexes[1] += 1
            current_indexes[0] += 1

        self.indexes.pop()


class ExecutionContext:
    __slots__ = ("event", "lifecycle_refs", "dispatchers")

    lifecycle_refs: Dict[str, List[Callable]]
    dispatchers: List[T_Dispatcher]

    def __init__(self, dispatchers: List[T_Dispatcher]) -> None:
        self.dispatchers = dispatchers

        self.lifecycle_refs = {i: [] for i in DEFAULT_LIFECYCLE_NAMES}


class ParameterContext:
    __slots__ = ("name", "annotation", "default", "dispatchers", "path")

    name: str
    annotation: Any
    default: Any

    dispatchers: List[T_Dispatcher]

    def __init__(self, name, annotation, default, dispatchers, using_path) -> None:
        self.name = name
        self.annotation = annotation
        self.default = default
        self.dispatchers = dispatchers
        self.path = DII_NestableIterable(using_path)

    def __repr__(self) -> str:
        return (
            "<ParameterContext name={0} annotation={1} default={2} locald={3}".format(
                self.name, self.annotation, self.default, self.dispatchers
            )
        )



from enum import Enum
from typing import TYPE_CHECKING, Any, List, Literal, Tuple, Union

if TYPE_CHECKING:
    from ..typing import T_Dispatcher


class TrackLogType(Enum):
    LookupStart = 0
    LookupEnd = 1

    Continue = 2
    Result = 3

    RequirementCrashed = 4


import sys

if sys.version_info >= (3, 8):
    T_TrackLogItem = Union[
        Tuple[Literal[TrackLogType.LookupStart], str, Any, Any],
        Tuple[Literal[TrackLogType.Continue], str, Any],
        Tuple[Literal[TrackLogType.Result], str, "T_Dispatcher"],
        Tuple[Literal[TrackLogType.LookupEnd], str],
        Tuple[Literal[TrackLogType.RequirementCrashed], str],
    ]
else:
    T_TrackLogItem = Union[
        Tuple[TrackLogType, str, Any, Any],
        Tuple[TrackLogType, str, Any],
        Tuple[TrackLogType, str, "T_Dispatcher"],
        Tuple[TrackLogType, str],
        Tuple[TrackLogType, str],
    ]


class TrackLog:
    __slots__ = ("log", "fluent_success")

    def __init__(self) -> None:
        self.log = []
        self.fluent_success = True

    log: List[T_TrackLogItem]
    fluent_success: bool



from functools import lru_cache
from itertools import chain
from typing import TYPE_CHECKING, Any, Dict, Generic, Iterable, List, TypeVar

from ..entities.context import ExecutionContext, ParameterContext
from ..entities.dispatcher import BaseDispatcher
from ..entities.event import Dispatchable
from ..entities.signatures import Force
from ..entities.track_log import TrackLog, TrackLogType
from ..exceptions import RequirementCrashed
from ..typing import (DEFAULT_LIFECYCLE_NAMES, T_Dispatcher,
                      T_Dispatcher_Callable)
from ..utilles import run_always_await_safely

if TYPE_CHECKING:
    from .. import Broadcast


class EmptyEvent(Dispatchable):
    class Dispatcher(BaseDispatcher):
        @staticmethod
        def catch(_):
            pass


T_Event = TypeVar("T_Event", bound=Dispatchable)

LIFECYCLE_ABS = {
    BaseDispatcher.beforeExecution,
    BaseDispatcher.afterDispatch,
    BaseDispatcher.afterExecution,
}


class DispatcherInterface(Generic[T_Event]):
    broadcast: "Broadcast"

    execution_contexts: List[ExecutionContext]
    parameter_contexts: List[ParameterContext]

    track_logs: List[TrackLog]

    @property
    def track_log(self) -> TrackLog:
        return self.track_logs[-1]

    def dispatcher_pure_generator(self) -> Iterable[T_Dispatcher]:
        return chain(
            self.execution_contexts[0].dispatchers,
            self.parameter_contexts[-1].dispatchers,
            self.execution_contexts[-1].dispatchers,
        )

    @staticmethod
    @lru_cache(maxsize=None)
    def get_lifecycle_refs(dispatcher: T_Dispatcher) -> Dict:
        result = {}

        for name in DEFAULT_LIFECYCLE_NAMES:
            v = getattr(dispatcher, name, None)
            if v and v.__func__ not in LIFECYCLE_ABS:
                result[name] = v
        return result

    def flush_lifecycle_refs(
        self,
        dispatchers: Iterable[T_Dispatcher],
    ):
        from graia.broadcast.entities.dispatcher import BaseDispatcher

        lifecycle_refs = self.execution_contexts[-1].lifecycle_refs

        for dispatcher in dispatchers:
            if dispatcher.__class__ is not type and not isinstance(
                dispatcher, BaseDispatcher
            ):
                continue

            result = self.get_lifecycle_refs(dispatcher)
            if result:
                for k, v in result.items():
                    lifecycle_refs[k].append(v)

    async def exec_lifecycle(self, lifecycle_name: str, *args, **kwargs):
        lifecycle_funcs = self.execution_contexts[-1].lifecycle_refs.get(lifecycle_name)
        if lifecycle_funcs:
            for func in lifecycle_funcs:
                await run_always_await_safely(func, self, *args, **kwargs)

    def inject_local_raw(self, *dispatchers: T_Dispatcher):
        # 为什么没有 flush: 因为这里的 lifecycle 是无意义的, 不会被 exec.
        for dispatcher in dispatchers[::-1]:
            self.parameter_contexts[-1].dispatchers.insert(0, dispatcher)

    def inject_execution_raw(self, *dispatchers: T_Dispatcher):
        for dispatcher in dispatchers:
            self.execution_contexts[-1].dispatchers.insert(0, dispatcher)

        self.flush_lifecycle_refs(dispatchers)

    def inject_global_raw(self, *dispatchers: T_Dispatcher):
        # self.dispatchers.extend(dispatchers)
        for dispatcher in dispatchers[::-1]:
            self.execution_contexts[0].dispatchers.insert(1, dispatcher)

        self.flush_lifecycle_refs(dispatchers)

    @property
    def name(self) -> str:
        return self.parameter_contexts[-1].name

    @property
    def annotation(self) -> Any:
        return self.parameter_contexts[-1].annotation

    @property
    def default(self) -> Any:
        return self.parameter_contexts[-1].default

    @property
    def event(self) -> T_Event:
        return self.broadcast.event_ctx.get()  # type: ignore

    @property
    def global_dispatcher(self) -> List[T_Dispatcher]:
        return self.execution_contexts[0].dispatchers

    @property
    def current_path(self) -> Iterable[T_Dispatcher]:
        return self.parameter_contexts[-1].path

    @property
    def has_current_exec_context(self) -> bool:
        return len(self.execution_contexts) >= 2

    @property
    def has_current_param_context(self) -> bool:
        return len(self.parameter_contexts) >= 2

    def __init__(self, broadcast_instance: "Broadcast") -> None:
        self.broadcast = broadcast_instance
        self.execution_contexts = [ExecutionContext([])]
        self.parameter_contexts = [
            ParameterContext(
                None,
                None,
                None,
                [],
                [
                    [],
                ],
            )
        ]
        self.track_logs = [TrackLog()]

    def clean(self):
        self.execution_contexts.pop()
        self.track_logs.pop()

    def start_execution(
        self,
        dispatchers: List[T_Dispatcher],
        track_log_receiver: TrackLog,
    ):
        self.execution_contexts.append(ExecutionContext(dispatchers))
        self.track_logs.append(track_log_receiver)
        self.flush_lifecycle_refs(self.dispatcher_pure_generator())
        return self

    @staticmethod
    @lru_cache(maxsize=None)
    def dispatcher_callable_detector(dispatcher: T_Dispatcher) -> T_Dispatcher_Callable:
        return getattr(dispatcher, "catch", dispatcher)  # type: ignore

    def init_dispatch_path(self) -> List[List[T_Dispatcher]]:
        return [
            [],
            self.execution_contexts[0].dispatchers,
            self.parameter_contexts[-1].dispatchers,
            self.execution_contexts[-1].dispatchers,
        ]

    async def lookup_param_without_log(
        self,
        name: str,
        annotation: Any,
        default: Any,
        using_path: List[List[T_Dispatcher]] = None,
    ) -> Any:
        self.parameter_contexts.append(
            ParameterContext(
                name, annotation, default, [], using_path or self.init_dispatch_path()
            )
        )

        try:
            for dispatcher in self.current_path:
                result = await self.dispatcher_callable_detector(dispatcher)(self)

                if result is None:
                    continue

                if result.__class__ is Force:
                    return result.target

                return result
            else:
                raise RequirementCrashed(
                    "the dispatching requirement crashed: ",
                    self.name,
                    self.annotation,
                    self.default,
                )
        finally:
            self.parameter_contexts.pop()

    async def lookup_param(
        self,
        name: str,
        annotation: Any,
        default: Any,
        using_path: List[List[T_Dispatcher]] = None,
    ) -> Any:
        self.parameter_contexts.append(
            ParameterContext(
                name, annotation, default, [], using_path or self.init_dispatch_path()
            )
        )
        track_log = self.track_log.log

        track_log.append((TrackLogType.LookupStart, name, annotation, default))
        try:
            for dispatcher in self.current_path:
                result = await self.dispatcher_callable_detector(dispatcher)(self)

                if result is None:
                    self.track_log.fluent_success = False
                    track_log.append((TrackLogType.Continue, name, dispatcher))
                    continue

                if result.__class__ is Force:
                    return result.target

                track_log.append((TrackLogType.Result, name, dispatcher))
                return result
            else:
                track_log.append((TrackLogType.RequirementCrashed, name))
                raise RequirementCrashed(
                    "the dispatching requirement crashed: ",
                    self.name,
                    self.annotation,
                    self.default,
                )
        finally:
            track_log.append((TrackLogType.LookupEnd, name))
            self.parameter_contexts.pop()

    async def lookup_using_current(self) -> Any:
        for dispatcher in self.current_path:
            result = await self.dispatcher_callable_detector(dispatcher)(self)
            if result is None:
                continue

            if result.__class__ is Force:
                return result.target

            return result
        else:
            raise RequirementCrashed(
                "the dispatching requirement crashed: ",
                self.name,
                self.annotation,
                self.default,
            )

    async def lookup_by_directly(
        self,
        dispatcher: T_Dispatcher,
        name: str,
        annotation: Any,
        default: Any,
        using_path: List[List[T_Dispatcher]] = None,
    ) -> Any:
        self.parameter_contexts.append(
            ParameterContext(
                name, annotation, default, [], using_path or self.init_dispatch_path()
            )
        )

        dispatcher_callable = self.dispatcher_callable_detector(dispatcher)

        try:
            result = await dispatcher_callable(self)
            if result.__class__ is Force:
                return result.target

            return result
        finally:
            self.parameter_contexts.pop()



import asyncio
import sys
import traceback
from typing import Callable, Dict, Iterable, List, Optional, Set, Type, Union

from .builtin.event import ExceptionThrowed
from .entities.decorator import Decorator
from .entities.dispatcher import BaseDispatcher
from .entities.event import Dispatchable
from .entities.exectarget import ExecTarget
from .entities.listener import Listener
from .entities.namespace import Namespace
from .entities.signatures import Force, RemoveMe
from .entities.track_log import TrackLog, TrackLogType
from .exceptions import (DisabledNamespace, ExecutionStop, ExistedNamespace,
                         InvaildEventName, PropagationCancelled,
                         RegisteredEventListener, RequirementCrashed,
                         UnexistedNamespace)
from .interfaces.decorator import DecoratorInterface
from .interfaces.dispatcher import DispatcherInterface
from .typing import T_Dispatcher
from .utilles import (Ctx, argument_signature, dispatcher_mixin_handler,
                      group_dict, printer, run_always_await_safely)


class Broadcast:
    loop: asyncio.AbstractEventLoop

    default_namespace: Namespace
    namespaces: List[Namespace]
    listeners: List[Listener]

    dispatcher_interface: DispatcherInterface
    decorator_interface: DecoratorInterface

    event_ctx: Ctx[Dispatchable]

    debug_flag: bool

    def __init__(
        self,
        *,
        loop: asyncio.AbstractEventLoop = None,
        debug_flag: bool = False,
    ):
        self.loop = loop or asyncio.get_event_loop()
        self.default_namespace = Namespace(name="default", default=True)
        self.debug_flag = debug_flag
        self.namespaces = []
        self.listeners = []
        self.event_ctx = Ctx("bcc_event_ctx")
        self.dispatcher_interface = DispatcherInterface(self)
        self.decorator_interface = DecoratorInterface(self.dispatcher_interface)
        self.dispatcher_interface.execution_contexts[0].dispatchers.insert(
            0, self.decorator_interface
        )

        @self.dispatcher_interface.inject_global_raw
        async def _(interface: DispatcherInterface):
            if interface.annotation is interface.event.__class__:
                return interface.event
            elif interface.annotation is Broadcast:
                return interface.broadcast
            elif interface.annotation is DispatcherInterface:
                return interface

    def default_listener_generator(self, event_class) -> Iterable[Listener]:
        return list(
            filter(
                lambda x: all(
                    [
                        not x.namespace.hide,
                        not x.namespace.disabled,
                        event_class in x.listening_events,
                    ]
                ),
                self.listeners,
            )
        )

    async def layered_scheduler(
        self, listener_generator: Iterable[Listener], event: Dispatchable
    ):
        grouped: Dict[int, List[Listener]] = group_dict(
            listener_generator, lambda x: x.priority
        )
        event_dispatcher_mixin = dispatcher_mixin_handler(event.Dispatcher)
        with self.event_ctx.use(event):
            for _, current_group in sorted(grouped.items(), key=lambda x: x[0]):
                coros = [
                    self.Executor(target=i, dispatchers=event_dispatcher_mixin)
                    for i in current_group
                ]
                done_tasks, _ = await asyncio.wait(coros)
                for task in done_tasks:
                    if task.exception().__class__ is PropagationCancelled:
                        break

    async def Executor(
        self,
        target: Union[Callable, ExecTarget],
        dispatchers: List[T_Dispatcher] = None,
        post_exception_event: bool = True,
        print_exception: bool = True,
    ):
        is_exectarget = is_listener = False
        if isinstance(target, Listener):
            is_exectarget = is_listener = True
        elif isinstance(target, ExecTarget):
            is_exectarget = True
        event: Optional[Dispatchable] = self.event_ctx.get()

        if is_listener:
            if target.namespace.disabled:
                raise DisabledNamespace(
                    "catched a disabled namespace: {0}".format(target.namespace.name)
                )

        target_callable = target.callable if is_exectarget else target
        parameter_compile_result = {}

        dii = self.dispatcher_interface
        track_logs: TrackLog = TrackLog()
        self.dispatcher_interface.start_execution(
            [
                *(dispatchers or []),
                *(target.namespace.injected_dispatchers if is_listener else []),
                *(target.inline_dispatchers if is_exectarget else []),
            ],
            track_logs,
        )
        try:
            await dii.exec_lifecycle("beforeExecution")
            if is_exectarget:
                if target.maybe_failure:
                    initial_path = dii.init_dispatch_path()
                    for name, annotation, default in argument_signature(
                        target_callable
                    ):
                        target.param_paths.setdefault(name, initial_path)
                        parameter_compile_result[name] = await dii.lookup_param(
                            name, annotation, default, target.param_paths[name]
                        )
                else:
                    for name, annotation, default in argument_signature(
                        target_callable
                    ):
                        parameter_compile_result[
                            name
                        ] = await dii.lookup_param_without_log(
                            name, annotation, default, target.param_paths[name]
                        )
                for hl_d in target.decorators:
                    await dii.lookup_by_directly(
                        self.decorator_interface,
                        "_bcc_headless_decorators",
                        None,
                        hl_d,
                    )
            else:
                for name, annotation, default in argument_signature(target_callable):
                    parameter_compile_result[name] = await dii.lookup_param_without_log(
                        name, annotation, default, target.param_paths[name]
                    )
            await dii.exec_lifecycle("afterDispatch", None, None)
            result = await run_always_await_safely(
                target_callable, **parameter_compile_result
            )
        except (ExecutionStop, PropagationCancelled):
            raise
        except RequirementCrashed:
            traceback.print_exc()
            raise
        except Exception as e:
            if event is not None and event.__class__ is not ExceptionThrowed:
                if print_exception:
                    traceback.print_exc()
                elif post_exception_event:
                    self.postEvent(ExceptionThrowed(exception=e, event=event))
            raise
        finally:
            _, exception, tb = sys.exc_info()
            await dii.exec_lifecycle("afterExecution", exception, tb)
            dii.clean()
            if is_exectarget and not track_logs.fluent_success:
                paths = target.param_paths
                current_path: Optional[List["T_Dispatcher"]] = None
                has_failures = set()
                for log_type, *log in track_logs.log:
                    if log_type is TrackLogType.LookupStart:
                        current_path = paths[log[0]]
                    elif log_type is TrackLogType.Continue:
                        has_failures.add(log[0])
                    elif log_type is TrackLogType.Result:
                        current_path[0].append(log[1])
                    elif log_type is TrackLogType.LookupEnd:
                        current_path[0] = list(dict.fromkeys(current_path[0]))
                target.maybe_failure.symmetric_difference_update(has_failures)

        if result.__class__ is Force:
            return result.content
        elif result.__class__ is RemoveMe:
            if is_listener and target in self.listeners:
                self.listeners.remove(target)
        return result

    def postEvent(self, event: Dispatchable):
        self.loop.create_task(
            self.layered_scheduler(
                listener_generator=self.default_listener_generator(event.__class__),
                event=event,
            )
        )

    @staticmethod
    def event_class_generator(target=Dispatchable):
        for i in target.__subclasses__():
            yield i
            if i.__subclasses__():
                yield from Broadcast.event_class_generator(i)

    @staticmethod
    def findEvent(name: str):
        for i in Broadcast.event_class_generator():
            if i.__name__ == name:
                return i

    def getDefaultNamespace(self):
        return self.default_namespace

    def createNamespace(
        self, name, *, priority: int = 0, hide: bool = False, disabled: bool = False
    ):
        if self.containNamespace(name):
            raise ExistedNamespace(name, "has been created!")
        self.namespaces.append(
            Namespace(name=name, priority=priority, hide=hide, disabled=disabled)
        )
        return self.namespaces[-1]

    def removeNamespace(self, name):
        if self.containNamespace(name):
            for index, i in enumerate(self.namespaces):
                if i.name == name:
                    self.namespaces.pop(index)
                    return
        else:
            raise UnexistedNamespace(name)

    def containNamespace(self, name):
        for i in self.namespaces:
            if i.name == name:
                return True
        return False

    def getNamespace(self, name) -> "Namespace":
        if self.containNamespace(name):
            for i in self.namespaces:
                if i.name == name:
                    return i
            else:
                raise UnexistedNamespace(name)
        else:
            raise UnexistedNamespace(name)

    def hideNamespace(self, name):
        ns = self.getNamespace(name)
        ns.hide = True

    def unhideNamespace(self, name):
        ns = self.getNamespace(name)
        ns.hide = False

    def disableNamespace(self, name):
        ns = self.getNamespace(name)
        ns.disabled = True

    def enableNamespace(self, name):
        ns = self.getNamespace(name)
        ns.disabled = False

    def containListener(self, target):
        for i in self.listeners:
            if i.callable == target:
                return True
        return False

    def getListener(self, target):
        for i in self.listeners:
            if i.callable == target:
                return i

    def removeListener(self, target):
        self.listeners.remove(target)

    def receiver(
        self,
        event: Union[str, Type[Dispatchable]],
        priority: int = 16,
        dispatchers: List[T_Dispatcher] = None,
        namespace: Namespace = None,
        decorators: List[Decorator] = None,
    ):
        if isinstance(event, str):
            _name = event
            event = self.findEvent(event)
            if not event:
                raise InvaildEventName(_name + " is not vaild!")  # type: ignore
        priority = int(priority)

        def receiver_wrapper(callable_target):
            may_listener = self.getListener(callable_target)
            if not may_listener:
                self.listeners.append(
                    Listener(
                        callable=callable_target,
                        namespace=namespace or self.getDefaultNamespace(),
                        inline_dispatchers=dispatchers,
                        priority=priority,
                        listening_events=[event],
                        decorators=decorators,
                    )
                )
            else:
                if event not in may_listener.listening_events:
                    may_listener.listening_events.append(event)
                else:
                    raise RegisteredEventListener(
                        event.__name__, "has been registered!"
                    )
            return callable_target

        return receiver_wrapper



import inspect
from typing import Callable, Dict, List, Set, Union

from ..typing import T_Dispatcher
from .decorator import Decorator


class ExecTarget:
    callable: Callable
    inline_dispatchers: List[T_Dispatcher]
    decorators: List[Decorator]

    param_paths: Dict[str, List[List[T_Dispatcher]]]
    maybe_failure: Set[str]

    def __init__(
        self,
        callable: Callable,
        inline_dispatchers: List[T_Dispatcher] = None,
        decorators: List[Decorator] = None,
    ):
        self.callable = callable
        self.inline_dispatchers = inline_dispatchers or []
        self.decorators = decorators or []

        self.param_paths = {}
        self.maybe_failure = set(inspect.signature(self.callable).parameters.keys())
